# inspired by https://gitlab.mpi-sws.org/FP/iris-ci/raw/40b71102efe051c777a785035aebcbaa4c2ec19f/prepare-opam.sh
## This script installs the build dependencies for CI builds.

# Prepare OPAM configuration
export OPAMROOT="$(pwd)/opamroot"
export OPAMJOBS="$((2*$CPU_CORES))"
export OPAM_EDITOR="$(which false)"

# Make sure we got a good OPAM.
if test -d "$OPAMROOT"; then
    echo_color "$BOLDGREEN" "[prepare-opam] Re-using cached opam root"
else
    echo_color "$BOLDYELLOW" "[prepare-opam] Creating new opam root"
    mkdir "$OPAMROOT"
    opam init --no-setup -y
    FRESH_OPAM=yes
fi
eval `opam conf env`

if [[ -z "$FRESH_OPAM" ]]; then # skip if this is a fresh opam root
    # Update repositories
    opam update
fi

# Make sure we got the right set of repositories registered
if echo "$@" | egrep "(dev|beta)" > /dev/null; then
    # We are compiling against a dev version of something.  Get ourselves the dev repositories.
    test -d "$OPAMROOT/repo/coq-extra-dev" || opam repo add coq-extra-dev https://coq.inria.fr/opam/extra-dev -p 0
    test -d "$OPAMROOT/repo/coq-core-dev" || opam repo add coq-core-dev https://coq.inria.fr/opam/core-dev -p 5
else
    # No dev version, make sure we do not have the dev repositories.
    test -d "$OPAMROOT/repo/coq-extra-dev" && opam repo remove coq-extra-dev
    test -d "$OPAMROOT/repo/coq-core-dev" && opam repo remove coq-core-dev
fi
test -d "$OPAMROOT/repo/coq-released" || opam repo add coq-released https://coq.inria.fr/opam/released -p 10
test -d "$OPAMROOT/repo/iris-dev" || opam repo add iris-dev https://gitlab.mpi-sws.org/FP/opam-dev.git -p 20
echo

# We really want to run all of the following in one opam transaction, but due to opam limitations,
# that is not currently possible.  Once we use opam 2 for CI, we can first do all the pinning (with `-n`)
# followed by `opam upgrade --all build-dep/` (and hope this will also --fixup).

# Install fixed versions of some dependencies.
echo_color "$BOLDGREEN" "[prepare-opam] Processing pins"
while (( "$#" )); do # while there are arguments left
    PACKAGE="$1" ; shift
    KIND="$1" ; shift
    VERSION="$1" ; shift

    # Check if the pin is already set
    read -a PIN <<< $(opam pin list | (egrep "^$PACKAGE[. ]"))
    if [[ "${PIN[1]}" == "$KIND" && "${PIN[2]}" == "$VERSION" ]]; then
        echo_color "$BOLDGREEN" "[prepare-opam] $PACKAGE already $KIND-pinned to $VERSION"
    else
        echo_color "$BOLDYELLOW" "[prepare-opam] $KIND-pinning $PACKAGE to $VERSION"
        opam pin add -y -k "$KIND" "$PACKAGE" "$VERSION"
    fi
done
echo

if [[ -z "$FRESH_OPAM" ]]; then # skip if this is a fresh opam root
    # Upgrade cached things
    echo_color "$BOLDGREEN" "[prepare-opam] Upgrading packages"
    opam upgrade -y --fixup && opam upgrade -y
    echo
fi

# done
echo
coqc -v
