(** * Classical Lemmas, where the Proofs use tactics *)

Require Import GQM.Base.
Require Import GQM.GQM.
Require Import GQM.GQMDeduction.
Require Export GQM.GQMClassical.
Require Export GQM.GQMDedSubst.
Require Export GQM.GQMTactics.

Lemma gqm_idEquiv : forall A, |-- A <<->> A.
Proof.
  hSplit; apply: gqm_id.
Qed.

Lemma gqm_contrapEquiv : forall A B, |-- (A ->> B) <<->> (-! B ->> -! A).
Proof.
  move => A B.
  hSplit.
  - apply: gqm_contrap.
  - apply: gqm_I3.
Qed.

Lemma gqm_contrapEquiv2 : forall A B, |-- (-! A ->> B) <<->> (-!B ->> A).
Proof.
  move => A B.
  hSplit.
  - apply: gqm_contrap2.
  - apply: gqm_contrap2.
Qed.

Lemma gqm_contrapEquiv3 : forall A B, |-- (A ->> -! B) <<->> (B ->> -! A).
Proof.
  move => A B.
  hSplit.
  - apply: gqm_contrap3.
  - apply: gqm_contrap3.
Qed.

Lemma gqm_andCom : forall A B, |-- A & B <<->> B & A.
Proof.
  move => A B.
  hSplit; rewrite /And;
    apply: (tac_rewrite (fun X => _ ->> -! X) (gqm_contrapEquiv3 _ _));
    by apply: gqm_id.
Qed.

Lemma gqm_impSwapEquiv : forall A B C, |-- (A ->> B ->> C) <<->> (B ->> A ->> C).
Proof.
  move => A B C.
  by hSplit; eauto with GQMDB.
Qed.

Lemma gqm_andImpEquiv : forall A B C, |-- ((A & B) ->> C) <<->> (A ->> B ->> C).
Proof.
  move => A B C.
  hSplit.
  - apply: gqm_mp; first by apply: gqm_swap.
    apply: gqm_mp; last by apply: (gqm_andI A B).
    apply: gqm_mp; first by apply: gqm_chain.
    apply: gqm_chain2.
  - apply: gqm_mp; first by apply: gqm_swap.
    apply: gqm_mp; last by apply: (gqm_andE2 A B).
    apply: gqm_mp; first by apply: gqm_I2.
    apply: gqm_mp; last by apply: (gqm_andE1 A B).
    apply: gqm_mp; first by apply: gqm_chain.
    hRewrite (gqm_impSwapEquiv B (A ->> B ->> C) C) (fun X => A ->> X).
    apply: gqm_mp; first by apply: gqm_swap.
    apply: gqm_id.
Qed.

Lemma gqm_contrapAndR : forall A B C, |-- ((A & B) ->> C) ->> (A & -! C) ->> -! B.
Proof.
  move => A B C.
  hRewrite (gqm_andImpEquiv A B C) (fun X => X ->> A & -! C ->> -! B).
  hRewrite (gqm_andImpEquiv A (-!C) (-!B)) (fun X => (A ->> B ->> C) ->>X).
  apply: gqm_mp; first by apply: gqm_chain.
  apply: gqm_contrap.
Qed.

Lemma gqm_botAllEquiv : |-- |A| Bot <<->> Bot.
Proof.
  hSplit; last by apply: gqm_negE.
  apply: (gqm_AE _ (Var 0)).
Qed.

Lemma gqm_equivNegEquiv : forall A B, |-- (A <<->> B) <<->> (-! A <<->> -! B).
Proof.
  move => A B.
  hSplit.
  - rewrite /Equiv.
    apply: (tac_rewriteR (fun X => _ ->> X & _) (gqm_contrapEquiv _ _)).
    apply: (tac_rewriteR (fun X => _ ->> _ & X) (gqm_contrapEquiv _ _)).
    apply: (tac_rewrite (fun X => _ ->> X) (gqm_andCom _ _)).
    apply: gqm_id.
  - rewrite /Equiv.
    apply: (tac_rewrite (fun X => _ ->> X & _) (gqm_contrapEquiv _ _)).
    apply: (tac_rewrite (fun X => _ ->> _ & X) (gqm_contrapEquiv _ _)).
    apply: (tac_rewrite (fun X => _ ->> X) (gqm_andCom _ _)).
    apply: gqm_id.
Qed.

Lemma gqm_equivNegEquiv2 : forall A B, |-- (A <<->> -! B) <<->> (-! A <<->> B).
Proof.
  move => A B.
  hSplit.
  - rewrite /Equiv.
    apply: (tac_rewriteR (fun X => _ ->> X & _) (gqm_contrapEquiv2 _ _)).
    apply: (tac_rewriteR (fun X => _ ->> _ & X) (gqm_contrapEquiv3 _ _)).
    apply: (tac_rewrite (fun X => _ ->> X) (gqm_andCom _ _)).
    apply: gqm_id.
  - rewrite /Equiv.
    apply: (tac_rewrite (fun X => _ ->> X & _) (gqm_contrapEquiv3 _ _)).
    apply: (tac_rewrite (fun X => _ ->> _ & X) (gqm_contrapEquiv2 _ _)).
    apply: (tac_rewrite (fun X => _ ->> X) (gqm_andCom _ _)).
    apply: gqm_id.
Qed.

Lemma gqm_demorganAnd : forall A B, |-- A & B <<->> -! (-! A ||| -! B).
Proof.
  move => A B.
  rewrite /And /Or.
  apply: (tac_rewrite (fun X => -! (X ->> _) <<->> _) (gqm_dnegEquiv _)).
  apply: gqm_idEquiv.
Qed.

Lemma gqm_demorganOr : forall A B, |-- A ||| B <<->> -! (-! A & -! B).
Proof.
  move => A B.
  rewrite /And /Or.
  apply: (tac_rewriteR (fun X => _ <<->> X) (gqm_dnegEquiv _)).
  apply: (tac_rewriteR (fun X => _ <<->> _ ->> X) (gqm_dnegEquiv _)).
  apply: gqm_idEquiv.
Qed.